/*
This is the keyboard library for the linux serial terminal.
It scans each row for pressed keys, stores them in a 2d array as 0 or 1 and
matches this array to a character table.

The keyboard_poll_new_character function initializes gpio at first call,
manages debounce, long press delay and repeated output during long key press.
*/

#include <string.h>
#include "pico/stdlib.h"
#include "hardware/gpio.h"

/* polling and long press timers */
#define KEYBOARD_POLL_INTERVAL_MS 50
#define KEYBOARD_REPEAT_DELAY_MS 600
#define KEYBOARD_REPEAT_INTERVAL_MS 100

/* pin numbers */
static const uint16_t ROW_PINS[] = {8, 10, 12, 13, 14, 15};
static const uint16_t COLUMN_PINS[] = {9, 11, 28, 27, 26, 22, 21, 20, 19, 18, 17, 16};

/* caps lock */
bool caps_lock = false;

/* character tables */
static const uint8_t LOWER_CASE_TABLE[6][12] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0x1B, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '?'},
	{'\t', 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '+'},
	{0, 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', '#', '\r'},
	{0, '<', 'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '-'},
	{0, 0, 0, 0, 0, 0, 0, ' ', 0, 0, '\b', 0}};

static const uint8_t UPPER_CASE_TABLE[6][12] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0x1B, '!', 0x22, 0, '$', '%', '&', '/', '(', ')', '=', 0x60},
	{'\t', 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '*'},
	{0, 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 0x27, '\r'},
	{0, '>', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', ';', ':', '_'},
	{0, 0, 0, 0, 0, 0, 0, ' ', 0, 0, '\b', 0}};

static const uint8_t ALT_GR_TABLE[6][12] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0x5E, 0, 0, 0, 0, 0, '{', '[', ']', '}', 0x5C},
	{0, '@', 0, 0, 0, 0, 0, 0, 0, 0, 0, '~'},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, '|', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};

static const uint8_t CTRL_TABLE[6][12] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0x11, 0x17, 0x05, 0x12, 0x14, 0x1A, 0x15, '\t', 0x0F, 0x10, 0},
	{0, 0x01, 0x13, 0x04, 0x06, 0x07, 0x08, '\n', 0x0B, 0x0c, 0, 0},
	{0, 0, 0x19, 0x18, 0x03, 0x16, 0x02, 0x0E, 0x0D, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};

static const uint8_t FN_TABLE[12][6] = {
	{0x1B, '[', '[', 'A', 0, 0},
	{0x1B, '[', '[', 'B', 0, 0},
	{0x1B, '[', '[', 'C', 0, 0},
	{0x1B, '[', '[', 'D', 0, 0},
	{0x1B, '[', '[', 'E', 0, 0},
	{0x1B, '[', '1', '7', '~', 0},
	{0x1B, '[', '1', '8', '~', 0},
	{0x1B, '[', '1', '9', '~', 0},
	{0x1B, '[', '2', '0', '~', 0},
	{0x1B, '[', '2', '1', '~', 0},
	{0x1B, '[', '2', '3', '~', 0},
	{0x1B, '[', '2', '4', '~', 0},
};

void keyboard_init_gpio()
/* pin initialisation */
{
	for (uint16_t i = 0; i < 6; i++)
	{
		gpio_init(ROW_PINS[i]);
		gpio_set_dir(ROW_PINS[i], GPIO_OUT);
		gpio_put(ROW_PINS[i], 0);
	}
	for (uint16_t i = 0; i < 12; i++)
	{
		gpio_init(COLUMN_PINS[i]);
		gpio_set_dir(COLUMN_PINS[i], GPIO_IN);
		gpio_pull_down(COLUMN_PINS[i]);
	}
}

uint16_t keyboard_scan_keys(uint16_t key_matrix[6][12])
/* This function scans the matrix keyboard for pressed keys.
The function parameter is a 2d array of the size of the keyboard.
The function populates the array with 1 for pressed key and 0 for unpressed.
It returns the number of pressed keys. */
{
	uint16_t len_keys_pressed = 0;

	for (uint16_t row = 0; row < 6; row++)
	{
		gpio_put(ROW_PINS[row], 1);
		sleep_us(1);
		for (uint16_t column = 0; column < 12; column++)
		{
			if (gpio_get(COLUMN_PINS[column]))
			{
				key_matrix[row][column] = 1;
				len_keys_pressed += 1;
			}
			else
			{
				key_matrix[row][column] = 0;
			}
		}
		gpio_put(ROW_PINS[row], 0);
	}
	return len_keys_pressed;
}

uint16_t keyboard_lookup_character(uint16_t len_keys_pressed, uint16_t key_matrix[6][12], uint8_t *char_list)
/* This function matches the array of pressed keys ("key_matrix") to a character table.
It adds the found characters to the provided string ("char_list") with \0 at the end.
It returns the length of the string without \0. */
{
	uint16_t alt_flag = 0;

	if (key_matrix[5][5])
	/* checking for alt key and removing it */
	{
		alt_flag = 1;
		key_matrix[5][5] = 0;
	}
	if (len_keys_pressed == 1 + alt_flag)
	/* 1 key remaining after removing alt */
	{
		if (!caps_lock)
		/* trying to match with LOWER_CASE_TABLE */
		{
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = LOWER_CASE_TABLE[row][column];
					}
				}
			}
		}
		else
		/* trying to match with UPPER_CASE_TABLE */ {
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = UPPER_CASE_TABLE[row][column];
					}
				}
			}
		}
		if (char_list[alt_flag])
		/* found character */
		{
			if (alt_flag)
			/* adding alt if it was pressed */
			{
				char_list[0] = 0x1B;
				char_list[2] = '\0';
				return 2;
			}
			else
			{
				char_list[1] = '\0';
				return 1;
			}
		}
		else
		/* no character found, looking for special keys */
		{
			if (key_matrix[3][0])
			/* up */
			{
				char_list[0] = 0x1B;
				char_list[1] = '[';
				char_list[2] = 'A';
				char_list[3] = '\0';
				return 3;
			}
			else if (key_matrix[4][0])
			/* down */
			{
				char_list[0] = 0x1B;
				char_list[1] = '[';
				char_list[2] = 'B';
				char_list[3] = '\0';
				return 3;
			}
			else if (key_matrix[5][1])
			/* left */
			{
				char_list[0] = 0x1B;
				char_list[1] = '[';
				char_list[2] = 'D';
				char_list[3] = '\0';
				return 3;
			}
			else if (key_matrix[5][2])
			/* right */
			{
				char_list[0] = 0x1B;
				char_list[1] = '[';
				char_list[2] = 'C';
				char_list[3] = '\0';
				return 3;
			}
			else if (key_matrix[5][11])
			/* delete */
			{
				char_list[0] = 0x1B;
				char_list[1] = '[';
				char_list[2] = '3';
				char_list[3] = '~';
				char_list[4] = '\0';
				return 4;
			}
			else
			{
				for (uint16_t i = 0; i < 12; i++)
				/* looking for function keys */
				{
					if (key_matrix[0][i])
					/* found a function key */
					{
						uint16_t n = 0;
						while (FN_TABLE[i][n])
						/* looping though the function key's character array until we hit a \0 */
						{
							char_list[n] = FN_TABLE[i][n];
							n += 1;
						}
						char_list[n + 1] = '\0';
						return n;
					}
				}
				/* loop ended without finding pressed function key, nothing to return */
				return 0;
			}
		}
	}
	else if (len_keys_pressed == 2 + alt_flag)
	/* 2 keys remaining after removing alt */
	{
		if (key_matrix[5][4] && key_matrix[5][8])
		/* both shift keys pressed toggles caps lock */
		{
			strcpy(char_list, "caps");
			return 4;
		}
		if (!caps_lock && (key_matrix[5][4] || key_matrix[5][8]))
		/* shift key is pressed, matching with UPPER_CASE_TABLE */
		{
			/* setting shift keys to 0, omitting this leads to some glitches */
			key_matrix[5][4] = 0;
			key_matrix[5][8] = 0;
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = UPPER_CASE_TABLE[row][column];
					}
				}
			}
		}
		else if (caps_lock && (key_matrix[5][4] || key_matrix[5][8]))
		/* caps lock is activ and shift key is pressed, matching with LOWER_CASE_TABLE */
		{
			key_matrix[5][4] = 0;
			key_matrix[5][8] = 0;
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = LOWER_CASE_TABLE[row][column];
					}
				}
			}
		}
		else if (key_matrix[5][6])
		/* ctrl is pressed, matching with CTRL_TABLE */
		{
			key_matrix[5][6] = 0;
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = CTRL_TABLE[row][column];
					}
				}
			}
		}
		else if (key_matrix[5][3] || key_matrix[5][9])
		/* alt gr is pressed, matching with ALT_GR_TABLE */
		{
			key_matrix[5][3] = 0;
			key_matrix[5][9] = 0;
			for (uint16_t row = 0; row < 6; row++)
			{
				for (uint16_t column = 0; column < 12; column++)
				{
					if (key_matrix[row][column])
					{
						char_list[alt_flag] = ALT_GR_TABLE[row][column];
					}
				}
			}
		}
		else
		/* nonsense combination */
		{
			return 0;
		}
		if (char_list[alt_flag])
		/* found a valid key combination */
		{
			if (alt_flag)
			/* adding 0x1B if alt is pressed */
			{
				char_list[0] = 0x1B;
				char_list[2] = '\0';
				return 2;
			}
			else
			{
				char_list[1] = '\0';
				return 1;
			}
		}
		else
		/* no character found */
		{
			return 0;
		}
	}
	else
	/* zero or more than two keys remaining after removing alt */
	{
		return 0;
	}
}

uint16_t keyboard_poll_new_character(uint8_t *character_output)
/* Main keyboard polling function.
It initializes gpio at first call, adds the keyboard output to the provided array "character_output"
and returns the length of the output. It returns 0, if it is called earlier than KEYBOARD_POLL_INTERVAL_MS,
this eliminates switch bouncing. It also returns 0, if it would provide the same character_output earlier
than KEYBOARD_REPEAT_DELAY_MS. If KEYBOARD_REPEAT_DELAY_MS has passed, it keeps providing the same
character_output every KEYBOARD_REPEAT_INTERVAL_MS until the output changes. */
{
	static bool init_flag;
	static bool is_repeating;
	static absolute_time_t timestamp_polling;
	static absolute_time_t timestamp_repeating;
	static char last_character[10]; // keeping track of the last character

	if (!init_flag)
	/* initializing gpio, setting first timestamp */
	{
		init_flag = true;
		keyboard_init_gpio();
		timestamp_polling = get_absolute_time();
		strcpy(last_character, "null"); // assigning a value to the last_character variable, omitting this leads to some bugs
	}
	if (absolute_time_diff_us(timestamp_polling, get_absolute_time()) > KEYBOARD_POLL_INTERVAL_MS * 1000)
	/* scanning keyboard only if KEYBOARD_POLL_INTERVAL_MS has passed */
	{
		timestamp_polling = get_absolute_time();
		uint16_t key_matrix[6][12];
		uint8_t new_character[10];
		uint16_t len_keys_pressed;
		uint16_t len_char_list;

		len_keys_pressed = keyboard_scan_keys(key_matrix);
		len_char_list = keyboard_lookup_character(len_keys_pressed, key_matrix, new_character);

		if (len_char_list)
		/* characters detected */
		{
			if (strcmp(last_character, new_character))
			/* different character than the last one */
			{
				if (!strcmp(new_character, "caps"))
				/* caps lock toggle, setting len_char_list to 0 means no printable character */
				{
					caps_lock = !caps_lock;
					len_char_list = 0;
				}
				strcpy(last_character, new_character);
				strcpy(character_output, new_character);
				is_repeating = false;
				timestamp_repeating = get_absolute_time();
				return len_char_list;
			}
			else if (!is_repeating && absolute_time_diff_us(timestamp_repeating, get_absolute_time()) > KEYBOARD_REPEAT_DELAY_MS * 1000)
			/* same character again, KEYBOARD_REPEAT_DELAY_MS has passed */
			{
				if (strcmp(new_character, "caps") == 0)
				{
					len_char_list = 0;
				}
				is_repeating = true;
				timestamp_repeating = get_absolute_time();
				return len_char_list; /* same character, no need to copy again */
			}
			else if (is_repeating && absolute_time_diff_us(timestamp_repeating, get_absolute_time()) > KEYBOARD_REPEAT_INTERVAL_MS * 1000)
			/* after KEYBOARD_REPEAT_DELAY_MS repeating output every KEYBOARD_REPEAT_INTERVAL_MS */
			{
				if (strcmp(new_character, "caps") == 0)
				{
					len_char_list = 0;
				}
				timestamp_repeating = get_absolute_time();
				return len_char_list;
			}
			else
			/* none of the conditions to provide an output are met */
			{
				return 0;
			}
		}
		else
		/* no character detected, putting an impossible result in last_character to avoid some bugs */
		{
			strcpy(last_character, "null");
			return 0;
		}
	}
	else
	/* KEYBOARD_POLL_INTERVAL_MS haven't passed yet, no keyboard scanning */
	{
		return 0;
	}
}
