// RA8875 library //

/*
Most of this is based on the Adafruit RA8875 library.
It has been stripped down specifically for a 800x480 display.
The function names have been changed to better reflect the
style of the Pico C SDK.

The original Adafruit library is BSD licensed:

Software License Agreement (BSD License)

Copyright (c) 2020, Limor Fried for Adafruit Industries
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holders nor the
names of its contributors may be used to endorse or promote products
derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <string.h>
#include <stdio.h>
#include "pico/stdlib.h"
#include "hardware/spi.h"
#include "hardware/gpio.h"
#include "ra8875_registers.h"

/* spi pins and parameters */
#define RA8875_SCK_PIN 2
#define RA8875_RX_PIN 4
#define RA8875_TX_PIN 3
#define RA8875_CSN_PIN 5
#define RA8875_RST_PIN 6
#define RA8875_SPI_BAUDRATE 10000000
#define SPI_ID spi0

/* global variable to hold the value of the MWCR0 register
using this results in simpler and shorter code */
uint8_t mwcr0_val = 0x00;

void ra8875_init_gpio()
/* spi initialization */
{
    spi_init(SPI_ID, RA8875_SPI_BAUDRATE);
    gpio_set_function(RA8875_RX_PIN, GPIO_FUNC_SPI);
    gpio_set_function(RA8875_SCK_PIN, GPIO_FUNC_SPI);
    gpio_set_function(RA8875_TX_PIN, GPIO_FUNC_SPI);
    gpio_init(RA8875_CSN_PIN);
    gpio_set_dir(RA8875_CSN_PIN, GPIO_OUT);
    gpio_put(RA8875_CSN_PIN, 1);
    gpio_init(RA8875_RST_PIN);
    gpio_set_dir(RA8875_RST_PIN, GPIO_OUT);
    gpio_put(RA8875_RST_PIN, 1);
}

void ra8875_write_command(uint8_t reg)
/* This function specifies a register to work with. */
{
    uint8_t buf[2];
    buf[0] = RA8875_CMDWRITE;
    buf[1] = reg;
    gpio_put(RA8875_CSN_PIN, 0);
    spi_write_blocking(SPI_ID, buf, 2);
    gpio_put(RA8875_CSN_PIN, 1);
}

void ra8875_write_data(uint8_t val)
/* This function writes data to a previously specified register. */
{
    uint8_t buf[2];
    buf[0] = RA8875_DATAWRITE;
    buf[1] = val;
    gpio_put(RA8875_CSN_PIN, 0);
    spi_write_blocking(SPI_ID, buf, 2);
    gpio_put(RA8875_CSN_PIN, 1);
}

void ra8875_write_2data(uint8_t val1, uint8_t val2)
/* This function writes data to a previously specified register. */
{
    uint8_t buf[4];
    buf[0] = RA8875_DATAWRITE;
    buf[1] = val1;
    buf[2] = RA8875_DATAWRITE;
    buf[3] = val2;
    gpio_put(RA8875_CSN_PIN, 0);
    spi_write_blocking(SPI_ID, buf, 4);
    gpio_put(RA8875_CSN_PIN, 1);
}

void ra8875_write_reg(uint8_t reg, uint8_t val)
/* This function writes a value to a register. */
{
    ra8875_write_command(reg);
    ra8875_write_data(val);
}

uint8_t ra8875_read_data()
/* This function reads the contents of a previously selected register. */
{
    uint8_t buf;
    uint8_t reg = RA8875_DATAREAD;
    gpio_put(RA8875_CSN_PIN, 0);
    spi_write_blocking(SPI_ID, &reg, 1);
    spi_read_blocking(SPI_ID, 0, &buf, 1);
    gpio_put(RA8875_CSN_PIN, 1);
    return buf;
}

void ra8875_blocking_wait(uint8_t regname, uint8_t waitflag)
/* This function polls the RA8875 for busy state. The possible registers
and waitflags are specified in the technical documentation. */
{
    uint8_t temp;
    do
    {
        ra8875_write_command(regname);
        temp = ra8875_read_data();
    } while (temp & waitflag);
}

void ra8875_init_display()
/* This function sets a lot of timing registers. I don't understand most of it. */
{
    const uint8_t pixclk = RA8875_PCSR_PDATL | RA8875_PCSR_2CLK;
    const uint8_t hsync_start = 32;
    const uint8_t hsync_pw = 96;
    const uint8_t hsync_nondisp = 26;
    const uint8_t vsync_pw = 2;
    const uint16_t vsync_nondisp = 32;
    const uint16_t vsync_start = 23;
    const uint16_t _width = 800;
    const uint16_t _height = 480;

    gpio_put(RA8875_RST_PIN, 0);
    sleep_ms(2);
    gpio_put(RA8875_RST_PIN, 1);
    sleep_ms(20);
    ra8875_write_reg(RA8875_PLLC1, RA8875_PWM_CLK_DIV2048);
    sleep_ms(1);
    ra8875_write_reg(RA8875_PLLC2, RA8875_PLLC2_DIV4);
    sleep_ms(1);
    ra8875_write_reg(RA8875_SYSR, RA8875_SYSR_16BPP | RA8875_SYSR_MCU8);
    ra8875_write_reg(RA8875_PCSR, pixclk);
    sleep_ms(1);
    ra8875_write_reg(RA8875_HDWR, (_width / 8) - 1);
    ra8875_write_reg(RA8875_HNDFTR, RA8875_HNDFTR_DE_HIGH);
    ra8875_write_reg(RA8875_HNDR, (hsync_nondisp - 2) / 8);
    ra8875_write_reg(RA8875_HSTR, hsync_start / 8 - 1);
    ra8875_write_reg(RA8875_HPWR, RA8875_HPWR_LOW + (hsync_pw / 8 - 1));
    ra8875_write_reg(RA8875_VDHR0, (_height - 1) & 0xFF);
    ra8875_write_reg(RA8875_VDHR1, (_height - 1) >> 8);
    ra8875_write_reg(RA8875_VNDR0, vsync_nondisp - 1);
    ra8875_write_reg(RA8875_VNDR1, vsync_nondisp >> 8);
    ra8875_write_reg(RA8875_VSTR0, vsync_start - 1);
    ra8875_write_reg(RA8875_VSTR1, vsync_start >> 8);
    ra8875_write_reg(RA8875_VPWR, RA8875_VPWR_LOW + vsync_pw - 1);
    ra8875_write_reg(RA8875_HSAW0, 0);
    ra8875_write_reg(RA8875_HSAW1, 0);
    ra8875_write_reg(RA8875_HEAW0, (_width - 1) & 0xFF);
    ra8875_write_reg(RA8875_HEAW1, (_width - 1) >> 8);
    ra8875_write_reg(RA8875_VSAW0, 0);
    ra8875_write_reg(RA8875_VSAW1, 0);
    ra8875_write_reg(RA8875_VEAW0, (_height - 1) & 0xFF);
    ra8875_write_reg(RA8875_VEAW1, (_height - 1) >> 8);
    ra8875_write_reg(RA8875_MCLR, RA8875_MCLR_START | RA8875_MCLR_FULL);
    sleep_ms(50);
}

void ra8875_turn_on()
{
    ra8875_write_reg(RA8875_PWRR, RA8875_PWRR_NORMAL | RA8875_PWRR_DISPON);
}

void ra8875_enable_gpiox()
{
    ra8875_write_reg(RA8875_GPIOX, 1);
}

void ra8875_set_pwm_clock(uint8_t clock)
/* This function sets the backlight pwm frequency. */
{
    ra8875_write_reg(RA8875_P1CR, RA8875_P1CR_ENABLE | (clock & 0xF));
}

void ra8875_set_pwm_duty(uint8_t p)
/* This function sets the backlight pwm duty cycle. 
Setting it to anything else than 255 leads to an audible noise. */
{
    ra8875_write_reg(RA8875_P1DCR, p);
}

void ra8875_setmode_text()
/* Function activating text mode. This is required to print characters. */
{
    mwcr0_val |= RA8875_MWCR0_TXTMODE;
    ra8875_write_reg(RA8875_MWCR0, mwcr0_val);
}

void ra8875_setmode_graphic(void)
/* The graphic mode is necessary to fill the cgram with custom glyphs
or write directly to th ddram. */
{
    mwcr0_val &= ~RA8875_MWCR0_TXTMODE;
    ra8875_write_reg(RA8875_MWCR0, mwcr0_val);
}

void ra8875_fill_cgram(const uint8_t bitmap_array[256][16])
/* This function fills the character graphic ram.
After specifying the cgram address, one glyph can be written in 16 continuous data write cycles.
The function takes an array containing 256 glyphs and loads them to the cgram */
{
    ra8875_setmode_graphic();
    ra8875_write_reg(RA8875_FNCR0, 0x00);
    ra8875_write_reg(RA8875_MWCR1, RA8875_MWCR1_CGRAM);
    for (uint16_t char_num = 0; char_num < 256; char_num++)
    {
        ra8875_write_reg(RA8875_CGSR, (uint8_t)char_num);
        ra8875_write_command(RA8875_MRWC);
        for (uint8_t i = 0; i < 16; i++)
        {
            ra8875_write_data(bitmap_array[char_num][i]);
        }
    }
    ra8875_write_reg(RA8875_MWCR1, 0x00);
    ra8875_setmode_text();
}

/* The next two functions set the source of the glyphs printed on the screen to
cgrom or cgram */
void ra8875_use_cgrom()
{
    ra8875_write_reg(RA8875_FNCR0, 0x00);
}

void ra8875_use_cgram()
{
    ra8875_write_reg(RA8875_FNCR0, 0x80);
}

void ra8875_cursor_on()
/* This function turns the visible cursor on, sets it to 8x16 pixels
and activates blinking. */
{
    mwcr0_val |= RA8875_MWCR0_CURSOR;
    mwcr0_val |= RA8875_MWCR0_BLINK;
    ra8875_write_reg(RA8875_MWCR0, mwcr0_val);
    ra8875_write_reg(RA8875_CURHS, 0x08);
    ra8875_write_reg(RA8875_CURVS, 0x0F);
    ra8875_write_reg(RA8875_BTCR, 0x10);
}

void ra8875_cursor_off()
/* turning cursor off */
{
    mwcr0_val &= ~RA8875_MWCR0_CURSOR;
    mwcr0_val &= ~RA8875_MWCR0_BLINK;
    ra8875_write_reg(RA8875_MWCR0, mwcr0_val);
}

void ra8875_set_cursor(uint16_t x, uint16_t y)
/* This function moves the cursor to the specified coordinates. */
{
    ra8875_write_reg(0x2A, x & 0xFF);
    ra8875_write_reg(0x2B, x >> 8);
    ra8875_write_reg(0x2C, y & 0xFF);
    ra8875_write_reg(0x2D, y >> 8);
}

/* The next two functions set the color registers */
void ra8875_set_font_color(uint16_t color)
{
    ra8875_write_reg(0x63, (color & 0xf800) >> 11);
    ra8875_write_reg(0x64, (color & 0x07e0) >> 5);
    ra8875_write_reg(0x65, color & 0x001f);
}

void ra8875_set_background_color(uint16_t color)
{
    ra8875_write_reg(0x60, (color & 0xf800) >> 11);
    ra8875_write_reg(0x61, (color & 0x07e0) >> 5);
    ra8875_write_reg(0x62, color & 0x001f);
}

void ra8875_write_text(char *buffer)
/* writing a string on the display */
{
    int len = strlen(buffer);
    ra8875_write_command(RA8875_MRWC);
    for (uint16_t i = 0; i < len; i++)
    {
        ra8875_write_data(buffer[i]);
        sleep_us(1);
    }
}

void ra8875_write_char(uint8_t buff)
/* writing a single character on the display */
{
    ra8875_write_reg(RA8875_MRWC, buff);
}

void ra8875_draw_rect(int16_t x1, int16_t y1, int16_t
 x2, int16_t y2, uint16_t color)
/* This function draws a rectangle on the display. This actually
works fine in text mode but it changes the font color */
{
    ra8875_write_reg(0x91, x1 & 0xFF);
    ra8875_write_reg(0x92, x1 >> 8);
    ra8875_write_reg(0x93, y1 & 0xFF);
    ra8875_write_reg(0x94, y1 >> 8);
    ra8875_write_reg(0x95, x2 & 0xFF);
    ra8875_write_reg(0x96, x2 >> 8);
    ra8875_write_reg(0x97, y2 & 0xFF);
    ra8875_write_reg(0x98, y2 >> 8);
    ra8875_write_reg(0x63, (color & 0xf800) >> 11);
    ra8875_write_reg(0x64, (color & 0x07e0) >> 5);
    ra8875_write_reg(0x65, color & 0x001f);
    ra8875_write_reg(RA8875_DCR, 0xB0);
    ra8875_blocking_wait(RA8875_DCR, RA8875_DCR_LINESQUTRI_STATUS);
}

/* The block transfer engine is used to directly manipulate contents of the RA8875 memory.
We will use it to copy parts of the screen to other parts of the screen. We need to set 
a starting pixel for source and a starting pixel for destination. Block size specifies
a rectangle. We have to set the direction of propagation of the copying process.
The function ra8875_bte_move_pos() goes from left to right and from top to bottom.
This means that the source and destination coordinates specify the top left corner of the block.
The function ra8875_bte_move_neg() is the opposite. In this case the source
and destination coordinates specify bottom right pixel of the block */

void ra8875_bte_set_source(uint16_t x, uint16_t y)
{
    ra8875_write_reg(RA8875_HSBE0, x & 0xFF);
    ra8875_write_reg(RA8875_HSBE1, x >> 8);
    ra8875_write_reg(RA8875_VSBE0, y & 0xFF);
    ra8875_write_reg(RA8875_VSBE1, y >> 8);
}

void ra8875_bte_set_destination(uint16_t x, uint16_t y)
{
    ra8875_write_reg(RA8875_HDBE0, x & 0xFF);
    ra8875_write_reg(RA8875_HDBE1, x >> 8);
    ra8875_write_reg(RA8875_VDBE0, y & 0xFF);
    ra8875_write_reg(RA8875_VDBE1, y >> 8);
}

void ra8875_bte_set_block_size(uint16_t w, uint16_t h)
{
    ra8875_write_reg(RA8875_BEWR0, w & 0xFF);
    ra8875_write_reg(RA8875_BEWR1, w >> 8);
    ra8875_write_reg(RA8875_BEHR0, h & 0xFF);
    ra8875_write_reg(RA8875_BEHR1, h >> 8);
}

void ra8875_bte_move_pos()
{
    ra8875_write_reg(RA8875_BECR1, 0xC2);
    ra8875_write_reg(RA8875_BECR0, 0x80);
    ra8875_blocking_wait(RA8875_BECR0, 0x80);
}

void ra8875_bte_move_neg()
{
    ra8875_write_reg(RA8875_BECR1, 0xC3);
    ra8875_write_reg(RA8875_BECR0, 0x80);
    ra8875_blocking_wait(RA8875_BECR0, 0x80);
}

void ra8875_set_ddram_address(uint16_t x, uint16_t y)
/* This function sets the memory write cursor to a specified pixel. */
{
    ra8875_write_reg(0x46, x & 0xFF);
    ra8875_write_reg(0x47, x >> 8);
    ra8875_write_reg(0x48, y & 0xFF);
    ra8875_write_reg(0x49, y >> 8);
}

void ra8875_set_active_window(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
/* Set the active window boudaries. Text and  ddram write operations
will keep auto-progressing and auto-wrapping within the active window. */
{
    ra8875_write_reg(0x30, x1 & 0xFF);
    ra8875_write_reg(0x31, x1 >> 8);
    ra8875_write_reg(0x32, y1 & 0xFF);
    ra8875_write_reg(0x33, y1 >> 8);
    ra8875_write_reg(0x34, x2 & 0xFF);
    ra8875_write_reg(0x35, x2 >> 8);
    ra8875_write_reg(0x36, y2 & 0xFF);
    ra8875_write_reg(0x37, y2 >> 8);
}

void ra8875_write_pixel(uint16_t val)
/* Set the color of a sinlge pixel */
{
    ra8875_setmode_graphic();
    ra8875_write_command(RA8875_MRWC);
    ra8875_write_data(val >> 8);
    ra8875_write_data(val & 0xFF);
    ra8875_setmode_text();
}

void ra8875_startup_wrapper()
/* This function calls all initialization steps of the RA8875. */
{
    ra8875_init_gpio();
    ra8875_init_display();
    ra8875_turn_on();
    ra8875_enable_gpiox();
    ra8875_set_pwm_clock(RA8875_PWM_CLK_DIV16);
    ra8875_set_pwm_duty(255);
    ra8875_setmode_text();
    ra8875_cursor_on();
}