/*
Serial terminal main binary

The program uses both rp2040 cores:

core_0
    turns led on
    initializes uart
    initializes the queue
    starts core 1
    polls uart rx and adds characters to the queue
    polls keyboard and writes keyboard output to uart tx

core 1
    the ra8875_startup_wrapper function initializes spi and ra8875
    polls queue for new characters
    runs the terminal interpreter managing output to the ra8875
*/

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "pico/util/queue.h"
#include "pico/multicore.h"
#include "ra8875.h"
#include "keyboard.h"
#include "box-bitmap.h"
#include "braille-bitmap.h"
#include "cyrillic-bitmap.h"

/* uart parameters */
#define UART_ID uart0
#define BAUD_RATE 115200
#define UART_TX_PIN 0
#define UART_RX_PIN 1
#define DATA_BITS 8
#define STOP_BITS 1
#define PARITY UART_PARITY_NONE

/* character buffer queue size */
#define BUFFER_SIZE 10000

/* screen and font dimensions
screen size is zero based, font size isn't
I know, its confusing */
#define SCREEN_X 799
#define SCREEN_Y 479
#define COLUMNS 99
#define ROWS 29
#define FONT_W 8
#define FONT_H 16

/* Cursor coordinates are zero based in rows and columns.
Scroll window is left inclusive and right exclusive. */
int cursor_x = 0;
int cursor_y = 0;
int cursor_x_saved = 0;
int cursor_y_saved = 0;
uint8_t scroll_window[2] = {0, ROWS + 1};

/* The RA8875 has a built in font covering ASCII and Latin-1 supplement.
The cgram of the RA8875 can be filled with 256 custom glyphs.
This program implements box, block, geometric and braille characters.
The glyphs are loaded dynamically, this variable keeps track of the loaded glyphs. */
enum charset_t
{
    box,
    braille,
    cyrillic
};
enum charset_t charset;

/* queue for buffering the characters from uart rx
and passing them between the cores. */
queue_t char_buffer_queue;

/* 3-bit color variables used in ANSI escape code interpreting functions.
They work as indexes for the 3-bit lookup table,
for example: COLORS_3BIT[color_brightness][font_color_3bit] gives the desired color. */
uint8_t font_color_3bit = 7;
uint8_t background_color_3bit = 0;
enum color_brightness_t
{
    dim = 0,
    normal = 1,
    bold = 2
};
enum color_brightness_t color_brightness = normal;

/* global variables holding the actual colors */
uint16_t font_color = 0xAD55;
uint16_t background_color = 0x0000;

/* This is an array of two pointers to the font and background color variables.
It allows an implementation of color inverting with a lot less code.
For font color: color_pointer[inverse_flag]
For background color: color_pointer[!inverse_flag] */
const uint16_t *color_pointer[2] = {&font_color, &background_color};
bool inverse_flag = false;

/* Lets put these in macros for better readability: */
#define FONT_COLOR *color_pointer[inverse_flag]
#define BACKGROUND_COLOR *color_pointer[!inverse_flag]

/* color lookup tables */
const uint16_t COLORS_3BIT[3][8] = {
    {0x0000, 0x5000, 0x02A0, 0x52A0, 0x0015, 0x5015, 0x02b5, 0x52AA},
    {0x0000, 0xA800, 0x0540, 0xAD40, 0x0015, 0xA815, 0x0555, 0xAD55},
    {0x52AA, 0xFAAA, 0x57EA, 0xFFEA, 0x52BF, 0xFABF, 0x57FF, 0xFFFF}};

const uint16_t COLORS_8BIT[256] = {
    0x0000, 0xA800, 0x0540, 0xAD40, 0x0015, 0xA815, 0x0555, 0xAD55,
    0x52AA, 0xFAAA, 0x57EA, 0xFFEA, 0x52BF, 0xFABF, 0x57FF, 0xFFFF,
    0x0000, 0x0006, 0x000C, 0x0013, 0x0019, 0x001F, 0x0180, 0x0186,
    0x018C, 0x0193, 0x0199, 0x019F, 0x0320, 0x0326, 0x032C, 0x0333,
    0x0339, 0x033F, 0x04C0, 0x04C6, 0x04CC, 0x04D3, 0x04D9, 0x04DF,
    0x0660, 0x0666, 0x066C, 0x0673, 0x0679, 0x067F, 0x07E0, 0x07E6,
    0x07EC, 0x07F3, 0x07F9, 0x07FF, 0x3000, 0x3006, 0x300C, 0x3013,
    0x3019, 0x301F, 0x3180, 0x3186, 0x318C, 0x3193, 0x3199, 0x319F,
    0x3320, 0x3326, 0x332C, 0x3333, 0x3339, 0x333F, 0x34C0, 0x34C6,
    0x34CC, 0x34D3, 0x34D9, 0x34DF, 0x3660, 0x3666, 0x366C, 0x3673,
    0x3679, 0x367F, 0x37E0, 0x37E6, 0x37EC, 0x37F3, 0x37F9, 0x37FF,
    0x6000, 0x6006, 0x600C, 0x6013, 0x6019, 0x601F, 0x6180, 0x6186,
    0x618C, 0x6193, 0x6199, 0x619F, 0x6320, 0x6326, 0x632C, 0x6333,
    0x6339, 0x633F, 0x64C0, 0x64C6, 0x64CC, 0x64D3, 0x64D9, 0x64DF,
    0x6660, 0x6666, 0x666C, 0x6673, 0x6679, 0x667F, 0x67E0, 0x67E6,
    0x67EC, 0x67F3, 0x67F9, 0x67FF, 0x9800, 0x9806, 0x980C, 0x9813,
    0x9819, 0x981F, 0x9980, 0x9986, 0x998C, 0x9993, 0x9999, 0x999F,
    0x9B20, 0x9B26, 0x9B2C, 0x9B33, 0x9B39, 0x9B3F, 0x9CC0, 0x9CC6,
    0x9CCC, 0x9CD3, 0x9CD9, 0x9CDF, 0x9E60, 0x9E66, 0x9E6C, 0x9E73,
    0x9E79, 0x9E7F, 0x9FE0, 0x9FE6, 0x9FEC, 0x9FF3, 0x9FF9, 0x9FFF,
    0xC800, 0xC806, 0xC80C, 0xC813, 0xC819, 0xC81F, 0xC980, 0xC986,
    0xC98C, 0xC993, 0xC999, 0xC99F, 0xCB20, 0xCB26, 0xCB2C, 0xCB33,
    0xCB39, 0xCB3F, 0xCCC0, 0xCCC6, 0xCCCC, 0xCCD3, 0xCCD9, 0xCCDF,
    0xCE60, 0xCE66, 0xCE6C, 0xCE73, 0xCE79, 0xCE7F, 0xCFE0, 0xCFE6,
    0xCFEC, 0xCFF3, 0xCFF9, 0xCFFF, 0xF800, 0xF806, 0xF80C, 0xF813,
    0xF819, 0xF81F, 0xF980, 0xF986, 0xF98C, 0xF993, 0xF999, 0xF99F,
    0xFB20, 0xFB26, 0xFB2C, 0xFB33, 0xFB39, 0xFB3F, 0xFCC0, 0xFCC6,
    0xFCCC, 0xFCD3, 0xFCD9, 0xFCDF, 0xFE60, 0xFE66, 0xFE6C, 0xFE73,
    0xFE79, 0xFE7F, 0xFFE0, 0xFFE6, 0xFFEC, 0xFFF3, 0xFFF9, 0xFFFF,
    0x0000, 0x0841, 0x10A2, 0x2104, 0x2965, 0x31A6, 0x4208, 0x4A69,
    0x5ACB, 0x630C, 0x6B6D, 0x7BCF, 0x8430, 0x8C71, 0x9CD3, 0xA534,
    0xB596, 0xBDD7, 0xC638, 0xD69A, 0xDEFB, 0xE73C, 0xF79E, 0xFFFF};

void uart_init_gpio()
/* uart initialisation */
{
    uart_init(UART_ID, BAUD_RATE);
    gpio_set_function(UART_TX_PIN, GPIO_FUNC_UART);
    gpio_set_function(UART_RX_PIN, GPIO_FUNC_UART);
    uart_set_hw_flow(UART_ID, false, false);
    uart_set_format(UART_ID, DATA_BITS, STOP_BITS, PARITY);
}

void led_on()
/* turn onboard led on */
{
    gpio_init(25);
    gpio_set_dir(25, GPIO_OUT);
    gpio_put(25, 1);
}

void update_cursor()
/* This function corrects off-screen cursor position
and sends the cursor position to the RA8875.
It uses some relational and logical operators to avoid a lot of branching. */
{
    cursor_x = (cursor_x >= 0 && cursor_x <= COLUMNS) * cursor_x + (cursor_x > COLUMNS) * COLUMNS;
    cursor_y = (cursor_y >= 0 && cursor_y <= ROWS) * cursor_y + (cursor_y > ROWS) * ROWS;
    ra8875_set_cursor(cursor_x * FONT_W, cursor_y * FONT_H);
}

void scroll_up(uint16_t lines)
/* Scrolls the scroll window upwards.
This function uses the ra8875 block transfer engine to copy content in the RA8875 memory. */
{
    ra8875_bte_set_source(0, (scroll_window[0] + lines) * FONT_H);
    ra8875_bte_set_destination(0, (scroll_window[0]) * FONT_H);
    ra8875_bte_set_block_size(SCREEN_X + 1, (scroll_window[1] - scroll_window[0] - lines) * FONT_H);
    ra8875_bte_move_pos();
    /* clear the area where the new lines are after copying */
    ra8875_draw_rect(0, (scroll_window[1] - lines) * FONT_H, SCREEN_X, scroll_window[1] * FONT_H - 1, background_color);
    /* The geometric drawing engine of the RA8875 uses the same color registers as the font drawing engine.
    We have to reset the font color after each call. */
    ra8875_set_font_color(FONT_COLOR);
}

void scroll_down(uint16_t lines)
/* scroll window downward scrolling */
{
    ra8875_bte_set_source(SCREEN_X, (scroll_window[1] - lines) * FONT_H - 1);
    ra8875_bte_set_destination(SCREEN_X, (scroll_window[1]) * FONT_H - 1);
    ra8875_bte_set_block_size(SCREEN_X + 1, (scroll_window[1] - scroll_window[0] - lines) * FONT_H);
    ra8875_bte_move_neg();
    ra8875_draw_rect(0, (scroll_window[0]) * FONT_H, SCREEN_X, (scroll_window[0] + lines) * FONT_H - 1, background_color);
    ra8875_set_font_color(FONT_COLOR);
}

void new_line()
/* This function moves cursor downward, updates the RA8875 cursor position
and scrolls screen if the cursor is at the last line of the scroll window.
The scroll window is right exclusive. */
{
    if (cursor_y == scroll_window[1] - 1)
    {
        scroll_up(1);
    }
    else
    {
        cursor_y += 1;
    }
    /* Actually, only after changing the cursor in the else block would calling
    update_cursor be absolutely nessecery, but we will assume that it is always called by this function */
    update_cursor();
}

void terminal_write_char(uint8_t input)
/* This function writes a character to display,
manages cursor advancing, line wrapping and scrolling.
It should be called only with printable characters, this way it
fixes the glitch when a newline character (\n) is present after the last column
and avoids unwanted empty lines. */
{
    if (cursor_x < 99)
    {
        ra8875_write_char(input);
        cursor_x += 1;
        update_cursor();
    }
    else if (cursor_x > 99)
    /* Line wrapping occurs only if function is called again.
    For this to work, the new_line function always has to call a cursor_update. */
    {
        cursor_x = 0;
        new_line();
        ra8875_write_char(input);
        cursor_x += 1;
        update_cursor();
    }
    else if (cursor_x == 99)
    /* Character is written at cursor_x=99,
    cursor_x stays at 100 (offscreen) and
    the RA8875 wraps the blinking cursor automatically to the next line. */
    {
        ra8875_write_char(input);
        cursor_x += 1;
    }
}

void short_ansi_interpreter(uint8_t *ansi)
/* interpreter function for ESC# sequences */
{
    switch (ansi[0])
    {
    case 'M':
        /* move cursor one line up, scrolling if needed */
        if (cursor_y == scroll_window[0])
        {
            scroll_down(1);
        }
        else
        {
            cursor_y -= 1;
            update_cursor();
        }
        break;
    case '7':
        /* save cursor position (DEC) */
        cursor_x_saved = cursor_x;
        cursor_y_saved = cursor_y;
        break;
    case '8':
        /* restore the cursor to the last saved position (DEC) */
        cursor_x = cursor_x_saved;
        cursor_y = cursor_y_saved;
        update_cursor();
        break;
    case 'c':
        /* reset */
        cursor_x = 0;
        cursor_y = 0;
        scroll_window[0] = 0;
        scroll_window[1] = ROWS + 1;
        font_color = 0xAD55;
        background_color = 0x0000;
        font_color_3bit = 7;
        background_color_3bit = 0;
        color_brightness = normal;
        inverse_flag = false;
        break;
    case 'D':
        /* linefeed, newline */
        new_line();
        break;
    case 'E':
        /* linefeed, newline */
        new_line();
        break;
    }
}

void ansi_cuu(uint16_t *ansi_parameters)
/* 'A' move cursor up # lines */
{
    cursor_y -= ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_cud(uint16_t *ansi_parameters)
/* 'B' move cursor down # lines */
{
    cursor_y += ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_cuf(uint16_t *ansi_parameters)
/* 'C' move cursor right # columns */
{
    cursor_x += ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_cub(uint16_t *ansi_parameters)
/* 'D' move cursor left # columns */
{
    cursor_x -= ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_cnl(uint16_t *ansi_parameters)
/* 'E' move cursor to beginning of next line, # lines down */
{
    cursor_x = 0;
    cursor_y += ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_cpl(uint16_t *ansi_parameters)
/* 'F' move cursor to beginning of previous line, # lines up */
{
    cursor_x = 0;
    cursor_y -= ansi_parameters[0] + !ansi_parameters[0];
    update_cursor();
}

void ansi_hpa(uint16_t *ansi_parameters)
/* 'G' move cursor to column # */
{
    cursor_x = ansi_parameters[0] - 1;
    update_cursor();
}

void ansi_vpa(uint16_t *ansi_parameters)
/* 'd' move cursor to row # */
{
    cursor_y = ansi_parameters[0] - 1;
    update_cursor();
}

void ansi_cup(uint16_t *ansi_parameters)
/* 'H' move the cursor to row n, column m */
{
    cursor_y = ansi_parameters[0] - 1;
    cursor_x = ansi_parameters[1] - 1;
    update_cursor();
}

void ansi_ed(uint16_t *ansi_parameters)
/* 'J' clear part of the screen */
{
    if (ansi_parameters[0] == 0)
    /* clear from cursor to end of the screen */
    {
        ra8875_draw_rect(cursor_x * FONT_W, cursor_y * FONT_H, SCREEN_X, (cursor_y + 1) * FONT_H - 1, background_color);
        ra8875_draw_rect(0, (cursor_y + 1) * FONT_H, SCREEN_X, SCREEN_Y, background_color);
    }
    else if (ansi_parameters[0] == 1)
    /* clear from cursor to beginning of the screen */
    {
        ra8875_draw_rect(0, cursor_y * FONT_H, cursor_x * FONT_W - 1, (cursor_y + 1) * FONT_H - 1, background_color);
        ra8875_draw_rect(0, 0, SCREEN_X, cursor_y * FONT_H - 1, background_color);
    }
    else if (ansi_parameters[0] == 2 || ansi_parameters[0] == 3)
    /* clear entire screen */
    {
        ra8875_draw_rect(0, 0, SCREEN_X, SCREEN_Y, background_color);
    }
    ra8875_set_font_color(FONT_COLOR);
}

void ansi_el(uint16_t *ansi_parameters)
/* erase part of the line */
{
    if (ansi_parameters[0] == 0)
    /* clear from cursor to the end of the line */
    {
        ra8875_draw_rect(cursor_x * FONT_W, cursor_y * FONT_H, SCREEN_X, (cursor_y + 1) * FONT_H - 1, background_color);
    }
    if (ansi_parameters[0] == 1)
    /* clear from cursor to beginning of the line */
    {
        ra8875_draw_rect(0, cursor_y * FONT_H, cursor_x * FONT_W - 1, (cursor_y + 1) * FONT_H - 1, background_color);
    }
    if (ansi_parameters[0] == 2)
    /* clear entire line */
    {
        ra8875_draw_rect(0, cursor_y * FONT_H, SCREEN_X, (cursor_y + 1) * FONT_H - 1, background_color);
    }
    ra8875_set_font_color(FONT_COLOR);
}

void ansi_ech(uint16_t *ansi_parameters)
/* 'X' erase the indicated # of characters on current line */
{
    uint16_t num = ansi_parameters[0] + !ansi_parameters[0];
    ra8875_draw_rect(cursor_x * FONT_W, cursor_y * FONT_H, (cursor_x + num) * FONT_W - 1, (cursor_y + 1) * FONT_H - 1, background_color);
    ra8875_set_font_color(FONT_COLOR);
}

void ansi_dch(uint16_t *ansi_parameters)
/* 'P' delete the indicated # of characters on current line, shift left to close gap */
{
    uint16_t num = ansi_parameters[0] + !ansi_parameters[0];
    ra8875_bte_set_source((cursor_x + num) * FONT_W, cursor_y * FONT_H);
    ra8875_bte_set_destination(cursor_x * FONT_W, cursor_y * FONT_H);
    ra8875_bte_set_block_size(SCREEN_X + 1 - (cursor_x + num) * FONT_W, FONT_H);
    ra8875_bte_move_pos();
    ra8875_draw_rect(SCREEN_X + 1 - num * FONT_W, cursor_y * FONT_H, SCREEN_X, (cursor_y + 1) * FONT_H - 1, background_color);
    ra8875_set_font_color(FONT_COLOR);
}

void ansi_dl(uint16_t *ansi_parameters)
/* 'M' delete the indicated # of lines, scroll to close the gap */
{
    uint16_t scroll_window_top = scroll_window[0];
    scroll_window[0] = cursor_y;
    scroll_up(ansi_parameters[0] + !ansi_parameters[0]);
    scroll_window[0] = scroll_window_top;
}

void ansi_su(uint16_t *ansi_parameters)
/* 'S' scroll whole page up by n (default 1) lines */
{
    scroll_up(ansi_parameters[0] + !ansi_parameters[0]);
}

void ansi_sd(uint16_t *ansi_parameters)
/* 'T' scroll whole page down by n (default 1) lines */
{
    scroll_down(ansi_parameters[0] + !ansi_parameters[0]);
}

void ansi_ich(uint16_t *ansi_parameters)
/* '@' insert # of blank characters */
{
    uint16_t num = ansi_parameters[0] + !ansi_parameters[0];
    ra8875_bte_set_source(SCREEN_X + 1 - num * FONT_W, (cursor_y + 1) * FONT_H - 1);
    ra8875_bte_set_destination(SCREEN_X + 1, (cursor_y + 1) * FONT_H - 1);
    ra8875_bte_set_block_size(SCREEN_X + 1 - (cursor_x + num) * FONT_W, FONT_H);
    ra8875_bte_move_neg();
    ra8875_draw_rect(cursor_x * FONT_W, cursor_y * FONT_H, (cursor_x + num) * FONT_W - 1, (cursor_y + 1) * FONT_H - 1, background_color);
    ra8875_set_font_color(FONT_COLOR);
}

void ansi_il(uint16_t *ansi_parameters)
/* 'L' insert the indicated # of blank lines */
{
    uint16_t scroll_window_top = scroll_window[0];
    scroll_window[0] = cursor_y;
    scroll_down(ansi_parameters[0] + !ansi_parameters[0]);
    scroll_window[0] = scroll_window_top;
}

void ansi_csr(uint16_t *ansi_parameters)
/* 'r' set scrolling region
The terminal scrolling window is left inclusive and right exclusive.
Line numbering is zero based -> 0 is first, 29 is last line. */
{
    scroll_window[0] = ansi_parameters[0] + !ansi_parameters[0] - 1;
    scroll_window[1] = ansi_parameters[1] + !ansi_parameters[1] * (ROWS + 1);
}

void ansi_civis(uint16_t *ansi_parameters, uint8_t len_ansi_parameters)
/* '?25l' make cursor invisible */
{
    if (ansi_parameters[len_ansi_parameters - 1] == 25)
    {
        ra8875_cursor_off();
    }
}

void ansi_cnorm(uint16_t *ansi_parameters, uint8_t len_ansi_parameters)
/* '?25h' make cursor appear normal */
{
    if (ansi_parameters[len_ansi_parameters - 1] == 25)
    {
        ra8875_cursor_on();
    }
}

void ansi_sgr(uint16_t *ansi_parameters, uint8_t len_ansi_parameters)
/* 'm' select graphic rendition
This can have a lot of numeric parameters and we will only know at the end
what the function schould actually change.
For example something like ESC[38;5;27;48;5;79m or ESC[0;1;29;37m.*/
{
    /* some flags to mark what to change */
    bool set_font_color = false;
    bool set_background_color = false;
    bool set_color_brightness = false;

    for (uint16_t i = 0; i < len_ansi_parameters; i++)
    /* This loop iterates through all numeric parameters and interprets them. */
    {
        if (ansi_parameters[i] == 38 && ansi_parameters[i + 1] == 5)
        /* font color 8-bit ESC[38;5;#m
        Tets the font color, advances the iterator to skip the interpreted parameters,
        sets set_font_color to false, because we already changed that.
        Omitting this step leads to some glitches. */
        {
            font_color = COLORS_8BIT[ansi_parameters[i + 2]];
            ra8875_set_font_color(FONT_COLOR);
            i += 2;
            set_font_color = false;
        }
        else if (ansi_parameters[i] == 48 && ansi_parameters[i + 1] == 5)
        /* background color 8-bit - same as above */
        {
            background_color = COLORS_8BIT[ansi_parameters[i + 2]];
            ra8875_set_background_color(BACKGROUND_COLOR);
            i += 2;
            set_background_color = false;
        }
        else if (ansi_parameters[i] > 29 && ansi_parameters[i] < 38)
        /* 3-bit font color
        Here we have to account for bold or dim mode,
        the actual color setting happens after the loop ends. */
        {
            font_color_3bit = ansi_parameters[i] - 30;
            set_font_color = true;
        }
        else if (ansi_parameters[i] > 39 && ansi_parameters[i] < 48)
        /* 3-bit background color - same as above */
        {
            background_color_3bit = ansi_parameters[i] - 40;
            set_background_color = true;
        }
        else if (ansi_parameters[i] > 89 && ansi_parameters[i] < 98)
        /* 3-bit bold font color - setting color, resetting flags,
        I think at this point you know the drill:) */
        {
            font_color = COLORS_3BIT[2][ansi_parameters[i] - 90];
            ra8875_set_font_color(FONT_COLOR);
            set_font_color = false;
            set_color_brightness = false;
        }
        else if (ansi_parameters[i] > 99 && ansi_parameters[i] < 108)
        /* 3-bit bold background color */
        {
            background_color = COLORS_3BIT[2][ansi_parameters[i] - 100];
            ra8875_set_background_color(BACKGROUND_COLOR);
            set_background_color = false;
            set_color_brightness = false;
        }
        else if (ansi_parameters[i] == 0)
        /* reset all modes (styles and colors) */
        {
            font_color = 0xAD55;
            background_color = 0x0000;
            font_color_3bit = 7;
            background_color_3bit = 0;
            color_brightness = normal;
            inverse_flag = false;
            ra8875_set_font_color(font_color);
            ra8875_set_background_color(background_color);
        }
        else if (ansi_parameters[i] == 1)
        /* set bold mode */
        {
            color_brightness = bold;
            set_color_brightness = true;
        }
        else if (ansi_parameters[i] == 2)
        /* set dim mode */
        {
            color_brightness = dim;
            set_color_brightness = true;
        }
        else if (ansi_parameters[i] == 7)
        /* set inverse mode */
        {
            inverse_flag = true;
            set_font_color = true;
            set_background_color = true;
        }
        else if (ansi_parameters[i] == 22)
        /* reset bold/dim */
        {
            color_brightness = normal;
            set_color_brightness = true;
        }
        else if (ansi_parameters[i] == 27)
        /* reset inverse mode */
        {
            inverse_flag = false;
            set_font_color = true;
            set_background_color = true;
        }
        else if (ansi_parameters[i] == 39)
        /* default font color */
        {
            font_color_3bit = 7;
            set_font_color = true;
        }
        else if (ansi_parameters[i] == 49)
        /* default background color */
        {
            background_color_3bit = 0;
            set_background_color = true;
        }
    }
    /* Interpreting reset flags:
    Often only ESC[1m is called to set the font to bold color.
    In this case only set_color_brightness is true, so we have to change set_font_color also to true. */
    set_font_color = set_font_color || (set_color_brightness && !set_background_color);

    /* This is the point, where we start changing colors. */
    if (set_font_color)
    {
        font_color = COLORS_3BIT[color_brightness][font_color_3bit];
        ra8875_set_font_color(FONT_COLOR);
    }
    if (set_background_color)
    /* This will always be normal brightness to avoid some graphical glitches. */
    {
        background_color = COLORS_3BIT[1][background_color_3bit];
        ra8875_set_background_color(BACKGROUND_COLOR);
    }
}

void terminal_draw_image(uint16_t *ansi_parameters)
/* This function writes bytes from the uart rx directly to the ddram of the ra8875. */
{
    uint8_t buf_r, buf_g, buf_b, bit0, bit1, new_lines;
    uint16_t active_window_x, active_window_y;

    /* Check and correct parameters */
    if (ansi_parameters[0] > SCREEN_X)
    {
        active_window_x = SCREEN_X + 1;
    }
    else
    {
        active_window_x = ansi_parameters[0];
    }
    if (ansi_parameters[1] > SCREEN_Y - FONT_H)
    {
        active_window_y = SCREEN_Y - FONT_H + 1;
    }
    else
    {
        active_window_y = ansi_parameters[1];
    }
    new_lines = (active_window_y / FONT_H) + 1;
    ra8875_cursor_off();

    /* making place for the image */
    if (cursor_y + new_lines >= scroll_window[1])
    {
        scroll_up(cursor_y + new_lines - scroll_window[1] + 1);
        cursor_y -= cursor_y + new_lines - scroll_window[1] + 1;
    }

    /* setting active window and first pixel loction */
    ra8875_set_ddram_address(0, cursor_y * FONT_H);
    ra8875_set_active_window(0, cursor_y * FONT_H, active_window_x - 1, cursor_y * FONT_H + active_window_y - 1);

    /* updating cursor loction */
    cursor_y += new_lines + 1;
    update_cursor();

    /* setting up memory write operation */
    ra8875_setmode_graphic();
    ra8875_write_command(0x02); // 0x02 -> MRWC

    /* start copying bytes from uart rx to ra8875 ddram */
    for (uint16_t y = 0; y < ansi_parameters[1]; y++)
    {
        for (uint16_t x = 0; x < ansi_parameters[0]; x++)
        {
            queue_remove_blocking(&char_buffer_queue, &buf_r);
            queue_remove_blocking(&char_buffer_queue, &buf_g);
            queue_remove_blocking(&char_buffer_queue, &buf_b);

            if (x < active_window_x && y < active_window_y)
            {
                bit0 = ((buf_r & 0x1F) << 3) | ((buf_g & 0x38) >> 3);
                bit1 = ((buf_g & 0x07) << 5) | (buf_b & 0x1F);
                ra8875_write_data(bit0);
                ra8875_write_data(bit1);
            }
        }
    }

    ra8875_set_active_window(0, 0, SCREEN_X, SCREEN_Y);
    ra8875_setmode_text();
    ra8875_cursor_on();
}

void ansi_interpreter(uint8_t last_char, uint16_t *ansi_parameters, uint8_t len_ansi_parameters)
/* This function is just a long switch to branch on the last
character of the control sequence. */
{
    switch (last_char)
    {
    case 'A':
        ansi_cuu(ansi_parameters);
        break;
    case 'B':
        ansi_cud(ansi_parameters);
        break;
    case 'C':
        ansi_cuf(ansi_parameters);
        break;
    case 'D':
        ansi_cub(ansi_parameters);
        break;
    case 'E':
        ansi_cnl(ansi_parameters);
        break;
    case 'F':
        ansi_cpl(ansi_parameters);
        break;
    case 'G':
        ansi_hpa(ansi_parameters);
        break;
    case 'd':
        ansi_vpa(ansi_parameters);
        break;
    case 'H':
        ansi_cup(ansi_parameters);
        break;
    case 'f':
        ansi_cup(ansi_parameters);
        break;
    case 'J':
        ansi_ed(ansi_parameters);
        break;
    case 'K':
        ansi_el(ansi_parameters);
        break;
    case 'X':
        ansi_ech(ansi_parameters);
        break;
    case 'P':
        ansi_dch(ansi_parameters);
        break;
    case 'M':
        ansi_dl(ansi_parameters);
        break;
    case 'S':
        ansi_su(ansi_parameters);
        break;
    case 'T':
        ansi_sd(ansi_parameters);
        break;
    case '@':
        ansi_ich(ansi_parameters);
        break;
    case 'L':
        ansi_il(ansi_parameters);
        break;
    case 'r':
        ansi_csr(ansi_parameters);
        break;
    case 'l':
        ansi_civis(ansi_parameters, len_ansi_parameters);
        break;
    case 'h':
        ansi_cnorm(ansi_parameters, len_ansi_parameters);
        break;
    case 'm':
        ansi_sgr(ansi_parameters, len_ansi_parameters);
        break;
    case 'z':
        terminal_draw_image(ansi_parameters);
        break;
    }
}

void ansi_grabber()
/* This function grabs the ANSI escape codes, 0x1B is omitted.
It converts the numeric parameters to an array of integers. */
{
    uint8_t ansi[30];                                              // this stores the escape code as char
    uint8_t len_ansi = 0;                                          // keeping track of the length
    uint16_t ansi_parameters[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // this array will be populated with integers, first we fill it with zeroes to avoid problems
    uint8_t len_ansi_parameters = 0;

    queue_remove_blocking(&char_buffer_queue, &ansi[0]);
    len_ansi += 1;

    if (ansi[0] == '[')
    /* '[' means that we have to keep collecting characters until a letter comes. */
    {
        do
        {
            queue_remove_blocking(&char_buffer_queue, &ansi[len_ansi]);
            len_ansi += 1;
        } while ((ansi[len_ansi - 1] >= '0' && ansi[len_ansi - 1] <= '9') || (ansi[len_ansi - 1] == ';') || (ansi[len_ansi - 1] == '?'));
        /* After texting the loop we need to interpret the contents of the ansi[] array. */
        if (len_ansi == 2)
        /* This means there are no numeric parameters, for example ESC[m.
        We need to pass some parameters to the next function, so lets assign values to them. */
        {
            ansi_parameters[0] = 0;
            len_ansi_parameters = 1;
        }
        else
        /* We have some numeric parameters, so we convert them to integers and add them to ansi_parameters. */
        {
            for (uint8_t i = 1; i < len_ansi; i++)
            /* this loop iterates over the contents of the ansi[] array */
            {
                if (ansi[i] >= '0' && ansi[i] <= '9')
                /* converting numeric characters to integer */
                {
                    ansi_parameters[len_ansi_parameters] = ansi_parameters[len_ansi_parameters] * 10 + (ansi[i] - '0');
                }
                else
                /* This can only be ';' separating numeric parameters (or '?'), so we move to the next position of the ansi_parameters[] array. */
                {
                    len_ansi_parameters += 1;
                }
            }
            /* If there wasnt any ';' in the escape code, we need to set len_ansi_parameters to 1. */
            len_ansi_parameters += !len_ansi_parameters;
        }
        /* passing the results to the next function */
        ansi_interpreter(ansi[len_ansi - 1], ansi_parameters, len_ansi_parameters);
    }
    else if (ansi[0] == '(' || ansi[0] == ')')
    /* These are not interpreted, so just skipping ahead in the queue. */
    {
        queue_remove_blocking(&char_buffer_queue, &ansi[1]);
    }
    else
    /* Not '(' ')' or '[' means the ANSI control sequence is probably just ESC#. */
    {
        short_ansi_interpreter(ansi);
    }
}

void terminal_entry_func(uint8_t buf)
/* Terminal interpreter entry function. It branches on the character received,
interprets control characters, like \n \r \t and manages a very limited Unicode support */
{
    if (buf == 0x1B)
    /* ANSI escape code */
    {
        ansi_grabber();
    }
    else if (buf == '\r')
    /* carriage return */
    {
        cursor_x = 0;
        update_cursor();
    }
    else if (buf == '\n')
    /* new line */
    {
        new_line();
    }
    else if (buf == '\b')
    /* backspace */
    {
        if (cursor_x)
        {
            cursor_x -= 1;
            update_cursor();
        }
        else if (cursor_y)
        /* If cursor_x is 0 an cursor_y is non-zero backspace wraps back to previous line. */
        {
            cursor_y -= 1;
            cursor_x = ROWS;
            update_cursor();
        }
    }
    else if (buf == '\t')
    /* tab */
    {
        cursor_x -= cursor_x % 8;
        cursor_x += 8;
        if (cursor_x > COLUMNS)
        {
            cursor_x = 0;
            new_line();
        }
        else
        {
            update_cursor();
        }
    }
    else if (buf < 0x80 && buf > 0x1F)
    /* printable ASCII character */
    {
        terminal_write_char(buf);
    }
    else if (buf >= 0xC0)
    /* Unicode character, so it can be 2-4 unit8_t long */
    {
        uint8_t buf1, buf2, buf3;

        queue_remove_blocking(&char_buffer_queue, &buf1);
        if (buf >= 0xE0)
        {
            queue_remove_blocking(&char_buffer_queue, &buf2);
            if (buf >= 0xF0)
            {
                queue_remove_blocking(&char_buffer_queue, &buf3);
            }
        }
        if (buf == 0xe2 && (0xa0 <= buf1) && (buf1 <= 0xa3))
        /* braille character */
        {
            if (charset != braille)
            /* Checking witch character set is loaded in the ra8875 cgram. */
            {
                ra8875_fill_cgram(BRAILLE_CHARACTER_BITMAPS);
                charset = braille;
            }
            ra8875_use_cgram();
            /* Remapping the UTF-8 encoding to the position of the character in cgram. */
            buf = ((buf1 - 0xa0) << 6) + (buf2 - 0x80);
            terminal_write_char(buf);
            ra8875_use_cgrom();
        }
        else if (buf == 0xe2 && (buf1 >= 0x94) && (buf1 <= 0x97))
        /* box/block/geometric character - same as above*/
        {
            if (charset != box)
            {
                ra8875_fill_cgram(BOX_CHARACTER_BITMAPS);
                charset = box;
            }
            ra8875_use_cgram();
            buf = ((buf1 - 0x94) << 6) + (buf2 - 0x80);
            terminal_write_char(buf);
            ra8875_use_cgrom();
        }
        else if (buf >= 0xd0 && buf <= 0xd3)
        /* cyrillic character */
        {
            if (charset != cyrillic)
            {
                ra8875_fill_cgram(CYRILLIC_CHARACTER_BITMAPS);
                charset = cyrillic;
            }
            ra8875_use_cgram();
            buf = ((buf - 0xd0) << 6) + (buf1 - 0x80);
            terminal_write_char(buf);
            ra8875_use_cgrom();
        }
        else if (buf == 0xc2 || buf == 0xc3)
        /* Latin-1 supplement characters are supported in the RA8875 cgrom, we only need to remap the UTF-8 encoding */
        {
            buf = ((buf - 0xc0) << 6) + (buf1 - 0x80);
            terminal_write_char(buf);
        }
        /* The following are some hand-picked Unicode characters. */
        else if (buf == 0xc5 && buf1 == 0x90)
        {
            terminal_write_char(0xd4);
        }
        else if (buf == 0xc5 && buf1 == 0x91)
        {
            terminal_write_char(0xf4);
        }
        else if (buf == 0xc5 && buf1 == 0xb0)
        {
            terminal_write_char(0xdb);
        }
        else if (buf == 0xc5 && buf1 == 0xb1)

        {
            terminal_write_char(0xfb);
        }
        else
        /* If the Unicode character is not supported, we still need to ptint something on the screen. */
        {
            terminal_write_char(0xB7);
        }
    }
}

void main_core1()
/* This is the core_1 entry function.
It manages display over spi and takes input from character buffer queue */
{
    uint8_t buf;

    ra8875_startup_wrapper();
    ra8875_set_font_color(font_color);
    ra8875_set_background_color(background_color);
    charset = box;
    ra8875_fill_cgram(BOX_CHARACTER_BITMAPS);

    while (1)
    {

        queue_remove_blocking(&char_buffer_queue, &buf);
        terminal_entry_func(buf);
    }
}

void main_core1_debug()
/* raw terminal output */
{
    uint8_t buf;

    ra8875_startup_wrapper();
    ra8875_set_background_color(background_color);
    ra8875_set_font_color(font_color);

    while (1)
    {
        queue_remove_blocking(&char_buffer_queue, &buf);
        terminal_write_char(buf);
    }
}

void main()
/* This is the core_0 main function.
It manages keyboard and uart polling.*/
{
    uint8_t keyboard_output[10];
    uint16_t len_keyboard_output;
    bool queue_full_flag;
    uint8_t char_dump;

    led_on();

    /* We have to initialize the queue */
    queue_init(&char_buffer_queue, sizeof(uint8_t), BUFFER_SIZE);

    /* If ESC is pressed start in debug mode. */
    gpio_init(9);
    gpio_set_dir(9, GPIO_IN);
    gpio_pull_down(9);
    gpio_init(10);
    gpio_set_dir(10, GPIO_OUT);
    gpio_put(10, 1);
    sleep_us(10);
    if (gpio_get(9))
    {
        multicore_launch_core1(main_core1_debug);
    }
    else
    {
        multicore_launch_core1(main_core1);
    }

    uart_init_gpio();

    while (1)
    {
        while (uart_is_readable(UART_ID))
        /* check the uart rx fifo */
        {
            uint8_t temp = uart_getc(UART_ID);

            /* Try to add the new character to the queue, check if it worked. */
            queue_full_flag = !queue_try_add(&char_buffer_queue, &temp);
            while (queue_full_flag)
            /* Keep trying until it works. */
            {
                queue_try_remove(&char_buffer_queue, &char_dump);
                queue_full_flag = !queue_try_add(&char_buffer_queue, &temp);
            }
        }
        /* If the uart rx fifo is empty, poll the keyboard for new character. */
        len_keyboard_output = keyboard_poll_new_character(keyboard_output);
        if (len_keyboard_output)
        {
            uart_write_blocking(UART_ID, keyboard_output, len_keyboard_output);
        }
    }
}